/* (C) Copyright 1984,85,86,87 Walter L. Peacock   All Rights Reserved    */
/*  C P Y R T B R O . C  7/25/84                             */
/* THIS PROGRAM BELONGS TO WALTER L. PEACOCK.  IT IS CONSIDERED A TRADE */
/* SECRET AND IS NOT TO BE DIVULGED OR USED BY PARTIES WHO HAVE NOT    */
/* RECEIVED WRITTEN AUTHORIZATION FROM THE OWNER.          */

/*---------------------------------------------------------------------------- 
 * cpyrtbro() will attempt to copy a cell from the right brother index 
 * node when the current index node underflows during a delete key.
 *---------------------------------------------------------------------------*/

#include "cbtree.h"

cpyrtbro(fdidx, pblk, btparmst)
int       fdidx;
BTIDXBLK *pblk;
BTC      *btparmst;
{
   extern char *calloc(), *strsave();
   extern long poptop();
   extern int strcmp(), strlen();
   extern void free();
   BTIDXBLK *broblk;
   long  lvlnode, idxnode, pos;
   char *endidxky;
   register int i;

   if (pblk->fwdpage < 2L)       /* there is no right brother   */
      return(-1);

   pos = (long) ((pblk->fwdpage - 1L) * btparmst->btidxlen);

   i = 0;     /* lock the record we are about to modify */
   while (lockunlk(fdidx, (long) pos, btparmst->btidxlen, LOCKR) != 1)
      if (++i > HDRTRYS)
         return(-1); 	/* i.e. unable to lock linked list record */

   /* place to read right brother index node */
   if( !(broblk = (BTIDXBLK *)calloc(
         sizeof(BTIDXBLK) + btparmst->btcells * SZBTIDX, SZCHAR) ))
      ckerror(- CK_NOMEM, "57");

   getidxr(fdidx, pblk->fwdpage, broblk, btparmst->btidxlen,   /* read it */
         btparmst->btkeylen);

   if (broblk->cellicnt <= (btparmst->btcells >> 1) ){   /* underflow  */
      lockunlk(fdidx, pos, btparmst->btidxlen, UNLOCKR);
      freekeys(broblk);
      FREE(broblk);
      return(-1);    /* can't copy anything from right brother */
   }

   idxnode = broblk->bwdpage;    /* point to original node */

   /* ok to copy first cell from right brother to our last postion */
   if (pblk->blkalloc > pblk->cellicnt)    /* memory already allocated  */
      FREE(pblk->btpage[pblk->cellicnt].skeynme);  /* don't need it */
   else
      ++pblk->blkalloc; /* we're adding an allocation from broblk */
   blkshft(pblk->btpage + pblk->cellicnt++, broblk->btpage, 1, SZBTIDX);

   /* compress the right brother by one cell */
   blkshft(broblk->btpage, broblk->btpage + 1, --broblk->blkalloc, SZBTIDX);
   --broblk->cellicnt;  /* we've deallocated a key from broblk */

   putidxr(fdidx, broblk->bwdpage, pblk, btparmst);
   putidxr(fdidx, pblk->fwdpage, broblk, btparmst);

   lockunlk(fdidx, pos, btparmst->btidxlen, UNLOCKR);

   /* now update tree: check the new end key against parent nodes */
   endidxky = strsave(pblk->btpage[pblk->cellicnt - 1].skeynme,
         btparmst->btkeylen);

   while( (lvlnode = poptop() ) != 0L)    /* node's parent */
   {
      getidxr(fdidx, lvlnode, broblk, btparmst->btidxlen, /* read parent */
            btparmst->btkeylen);

      i = 0;      /* find cell which pointed our index block */
      while (broblk->btpage[i++].btptr != idxnode){
         if (i >= broblk->cellicnt){
            lvlnode = broblk->fwdpage;
            getidxr(fdidx, lvlnode, broblk, btparmst->btidxlen,
                  btparmst->btkeylen);
            i = 0;
         }
      }
      i--;

      /* if the cell's key points BEFORE the new endkey, replace w endkey */
      if (strcmp(broblk->btpage[i].skeynme, endidxky) < 0){
         strcpy(broblk->btpage[i].skeynme, endidxky);
         putidxr(fdidx, lvlnode, broblk, btparmst);
      }
      /* now check parent's endkey */
      /* if parent's endkey points AFTER our endkey, revise our endkey */
      if (strcmp(broblk->btpage[broblk->cellicnt - 1].skeynme, endidxky) > 0)
         strcpy(endidxky, broblk->btpage[broblk->cellicnt - 1].skeynme);
      idxnode = lvlnode;

   }     /* countinue back up the tree */

   FREE(endidxky);
   freekeys(broblk);
   FREE(broblk);
   
   return(1);
}

